/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License;
 * you may not use this file except in compliance with the Elastic License.
 */
package org.elasticsearch.xpack.sql.plan.logical;

import java.util.Collections;
import java.util.List;
import java.util.Objects;

import org.elasticsearch.xpack.sql.expression.Attribute;
import org.elasticsearch.xpack.sql.tree.Location;

/**
 * A {@code UnaryPlan} is a {@code LogicalPlan} with exactly one child, for example, {@code WHERE x} in a
 * SQL statement is an {@code UnaryPlan}.
 */
public abstract class UnaryPlan extends LogicalPlan {

    private final LogicalPlan child;

    UnaryPlan(Location location, LogicalPlan child) {
        super(location, Collections.singletonList(child));
        this.child = child;
    }

    @Override
    public final UnaryPlan replaceChildren(List<LogicalPlan> newChildren) {
        if (newChildren.size() != 1) {
            throw new IllegalArgumentException("expected [1] child but received [" + newChildren.size() + "]");
        }
        return replaceChild(newChildren.get(0));
    }
    protected abstract UnaryPlan replaceChild(LogicalPlan newChild);

    public LogicalPlan child() {
        return child;
    }

    @Override
    public List<Attribute> output() {
        return child.output();
    }

    @Override
    public int hashCode() {
        return Objects.hashCode(child());
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null || getClass() != obj.getClass()) {
            return false;
        }

        UnaryPlan other = (UnaryPlan) obj;

        return Objects.equals(child, other.child);
    }
}
