/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License;
 * you may not use this file except in compliance with the Elastic License.
 */
package org.elasticsearch.xpack.sql.tree;

import org.elasticsearch.test.ESTestCase;

import static org.elasticsearch.xpack.sql.tree.LocationTests.randomLocation;

import java.util.Arrays;
import java.util.List;
import java.util.Objects;

import static java.util.Collections.emptyList;
import static java.util.Collections.singletonList;

public class NodeTests extends ESTestCase {
    public void testToString() {
        assertEquals("NoChildren[thing]", new NoChildren(randomLocation(), "thing").toString());
        {
            ChildrenAreAProperty empty = new ChildrenAreAProperty(randomLocation(), emptyList(), "thing");
            assertEquals("ChildrenAreAProperty[thing]", empty.toString());
            assertEquals("ChildrenAreAProperty[single]\n\\_ChildrenAreAProperty[thing]",
                new ChildrenAreAProperty(randomLocation(), singletonList(empty), "single").toString());
            assertEquals("ChildrenAreAProperty[many]\n"
                       + "|_ChildrenAreAProperty[thing]\n"
                      + "\\_ChildrenAreAProperty[thing]",
                new ChildrenAreAProperty(randomLocation(), Arrays.asList(empty, empty), "many").toString());
        }
        {
            NoChildren empty = new NoChildren(randomLocation(), "thing");
            assertEquals("AChildIsAProperty[single]\n"
                      + "\\_NoChildren[thing]",
                new AChildIsAProperty(randomLocation(), empty, "single").toString());
        }
    }

    public abstract static class Dummy extends Node<Dummy> {
        private final String thing;
        public Dummy(Location location, List<Dummy> children, String thing) {
            super(location, children);
            this.thing = thing;
        }

        public String thing() {
            return thing;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == this) {
                return true;
            }
            if (obj == null || obj.getClass() != getClass()) {
                return false;
            }
            Dummy other = (Dummy) obj;
            return thing.equals(other.thing)
                && children().equals(other.children());
        }

        @Override
        public int hashCode() {
            return Objects.hash(thing, children());
        }
    }

    public static class ChildrenAreAProperty extends Dummy {
        public ChildrenAreAProperty(Location location, List<Dummy> children, String thing) {
            super(location, children, thing);
        }

        @Override
        protected NodeInfo<ChildrenAreAProperty> info() {
            return NodeInfo.create(this, ChildrenAreAProperty::new, children(), thing());
        }

        @Override
        public ChildrenAreAProperty replaceChildren(List<Dummy> newChildren) {
            return new ChildrenAreAProperty(location(), newChildren, thing());
        }
    }

    public static class AChildIsAProperty extends Dummy {
        public AChildIsAProperty(Location location, Dummy child, String thing) {
            super(location, singletonList(child), thing);
        }

        @Override
        protected NodeInfo<AChildIsAProperty> info() {
            return NodeInfo.create(this, AChildIsAProperty::new, child(), thing());
        }

        @Override
        public AChildIsAProperty replaceChildren(List<Dummy> newChildren) {
            if (newChildren.size() != 1) {
                throw new IllegalArgumentException("expected [1] child but received [" + newChildren.size() + "]");
            }
            return new AChildIsAProperty(location(), newChildren.get(0), thing());
        }

        public Dummy child() {
            return children().get(0);
        }
    }

    public static class NoChildren extends Dummy {
        public NoChildren(Location location, String thing) {
            super(location, emptyList(), thing);
        }

        @Override
        protected NodeInfo<NoChildren> info() {
            return NodeInfo.create(this, NoChildren::new, thing());
        }

        @Override
        public Dummy replaceChildren(List<Dummy> newChildren) {
            throw new UnsupportedOperationException("no children to replace");
        }
    }
}
